package org.antlr.analysis;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.antlr.analysis.DFA;
import org.antlr.analysis.DFAState;
import org.antlr.analysis.Label;
import org.antlr.analysis.NFAConfiguration;
import org.antlr.analysis.NFAState;
import org.antlr.analysis.RuleClosureTransition;
import org.antlr.analysis.SemanticContext;
import org.antlr.analysis.State;
import org.antlr.analysis.Transition;
import org.antlr.misc.MultiMap;
import org.antlr.misc.Utils;
import org.antlr.runtime.Token;
import org.antlr.tool.ErrorManager;
import org.antlr.tool.Grammar;
import org.antlr.tool.GrammarAST;

public class DecisionProbe {
  public DFA dfa;
  
  protected Set<DFAState> statesWithSyntacticallyAmbiguousAltsSet = new HashSet<DFAState>();
  
  protected Map<DFAState, Set<Integer>> stateToSyntacticallyAmbiguousTokensRuleAltsMap = new HashMap<DFAState, Set<Integer>>();
  
  protected Set<DFAState> statesResolvedWithSemanticPredicatesSet = new HashSet<DFAState>();
  
  protected Map<DFAState, Map<Integer, SemanticContext>> stateToAltSetWithSemanticPredicatesMap = new HashMap<DFAState, Map<Integer, SemanticContext>>();
  
  protected Map<DFAState, Map<Integer, Set<Token>>> stateToIncompletelyCoveredAltsMap = new HashMap<DFAState, Map<Integer, Set<Token>>>();
  
  protected Set<DFAState> danglingStates = new HashSet<DFAState>();
  
  protected Set<Integer> altsWithProblem = new HashSet<Integer>();
  
  public boolean nonLLStarDecision = false;
  
  protected MultiMap<Integer, NFAConfiguration> stateToRecursionOverflowConfigurationsMap = new MultiMap<Integer, NFAConfiguration>();
  
  protected boolean timedOut = false;
  
  protected Map<Integer, Integer> stateReachable;
  
  public static final Integer REACHABLE_BUSY = Utils.integer(-1);
  
  public static final Integer REACHABLE_NO = Utils.integer(0);
  
  public static final Integer REACHABLE_YES = Utils.integer(1);
  
  protected Set<String> statesVisitedAtInputDepth;
  
  protected Set<Integer> statesVisitedDuringSampleSequence;
  
  public static boolean verbose = false;
  
  public DecisionProbe(DFA dfa) {
    this.dfa = dfa;
  }
  
  public String getDescription() {
    return this.dfa.getNFADecisionStartState().getDescription();
  }
  
  public boolean isReduced() {
    return this.dfa.isReduced();
  }
  
  public boolean isCyclic() {
    return this.dfa.isCyclic();
  }
  
  public boolean isDeterministic() {
    if (this.danglingStates.isEmpty() && this.statesWithSyntacticallyAmbiguousAltsSet.isEmpty() && this.dfa.getUnreachableAlts().isEmpty())
      return true; 
    if (this.statesWithSyntacticallyAmbiguousAltsSet.size() > 0) {
      for (DFAState d : this.statesWithSyntacticallyAmbiguousAltsSet) {
        if (!this.statesResolvedWithSemanticPredicatesSet.contains(d))
          return false; 
      } 
      return true;
    } 
    return false;
  }
  
  public boolean analysisOverflowed() {
    return (this.stateToRecursionOverflowConfigurationsMap.size() > 0);
  }
  
  public boolean isNonLLStarDecision() {
    return this.nonLLStarDecision;
  }
  
  public int getNumberOfStates() {
    return this.dfa.getNumberOfStates();
  }
  
  public List<Integer> getUnreachableAlts() {
    return this.dfa.getUnreachableAlts();
  }
  
  public Set<DFAState> getDanglingStates() {
    return this.danglingStates;
  }
  
  public Set<Integer> getNonDeterministicAlts() {
    return this.altsWithProblem;
  }
  
  public List<Integer> getNonDeterministicAltsForState(DFAState targetState) {
    Set<Integer> nondetAlts = targetState.getNonDeterministicAlts();
    if (nondetAlts == null)
      return null; 
    List<Integer> sorted = new LinkedList<Integer>();
    sorted.addAll(nondetAlts);
    Collections.sort(sorted);
    return sorted;
  }
  
  public Set<DFAState> getDFAStatesWithSyntacticallyAmbiguousAlts() {
    return this.statesWithSyntacticallyAmbiguousAltsSet;
  }
  
  public Set<Integer> getDisabledAlternatives(DFAState d) {
    return d.getDisabledAlternatives();
  }
  
  public void removeRecursiveOverflowState(DFAState d) {
    Integer stateI = Utils.integer(d.stateNumber);
    this.stateToRecursionOverflowConfigurationsMap.remove(stateI);
  }
  
  public List<Label> getSampleNonDeterministicInputSequence(DFAState targetState) {
    Set<DFAState> dfaStates = getDFAPathStatesToTarget(targetState);
    this.statesVisitedDuringSampleSequence = new HashSet<Integer>();
    List<Label> labels = new ArrayList<Label>();
    if (this.dfa == null || this.dfa.startState == null)
      return labels; 
    getSampleInputSequenceUsingStateSet(this.dfa.startState, targetState, dfaStates, labels);
    return labels;
  }
  
  public String getInputSequenceDisplay(List<? extends Label> labels) {
    Grammar g = this.dfa.nfa.grammar;
    StringBuilder buf = new StringBuilder();
    for (Iterator<? extends Label> it = labels.iterator(); it.hasNext(); ) {
      Label label = it.next();
      buf.append(label.toString(g));
      if (it.hasNext() && g.type != 1)
        buf.append(' '); 
    } 
    return buf.toString();
  }
  
  public List<? extends NFAState> getNFAPathStatesForAlt(int firstAlt, int alt, List<? extends Label> labels) {
    NFAState nfaStart = this.dfa.getNFADecisionStartState();
    List<NFAState> path = new LinkedList<NFAState>();
    for (int a = firstAlt; a <= alt; a++) {
      NFAState s = this.dfa.nfa.grammar.getNFAStateForAltOfDecision(nfaStart, a);
      path.add(s);
    } 
    NFAState altStart = this.dfa.nfa.grammar.getNFAStateForAltOfDecision(nfaStart, alt);
    NFAState isolatedAltStart = (NFAState)(altStart.transition[0]).target;
    path.add(isolatedAltStart);
    this.statesVisitedAtInputDepth = new HashSet<String>();
    getNFAPath(isolatedAltStart, 0, labels, path);
    return path;
  }
  
  public SemanticContext getSemanticContextForAlt(DFAState d, int alt) {
    Map<Integer, SemanticContext> altToPredMap = this.stateToAltSetWithSemanticPredicatesMap.get(d);
    if (altToPredMap == null)
      return null; 
    return altToPredMap.get(Utils.integer(alt));
  }
  
  public boolean hasPredicate() {
    return (this.stateToAltSetWithSemanticPredicatesMap.size() > 0);
  }
  
  public Set<DFAState> getNondeterministicStatesResolvedWithSemanticPredicate() {
    return this.statesResolvedWithSemanticPredicatesSet;
  }
  
  public Map<Integer, Set<Token>> getIncompletelyCoveredAlts(DFAState d) {
    return this.stateToIncompletelyCoveredAltsMap.get(d);
  }
  
  public void issueWarnings() {
    if (this.nonLLStarDecision && !this.dfa.getAutoBacktrackMode())
      ErrorManager.nonLLStarDecision(this); 
    issueRecursionWarnings();
    Set<DFAState> resolvedStates = getNondeterministicStatesResolvedWithSemanticPredicate();
    Set<DFAState> problemStates = getDFAStatesWithSyntacticallyAmbiguousAlts();
    if (problemStates.size() > 0) {
      Iterator<DFAState> it = problemStates.iterator();
      while (it.hasNext() && !this.dfa.nfa.grammar.NFAToDFAConversionExternallyAborted()) {
        DFAState d = it.next();
        Map<Integer, Set<Token>> insufficientAltToLocations = getIncompletelyCoveredAlts(d);
        if (insufficientAltToLocations != null && insufficientAltToLocations.size() > 0)
          ErrorManager.insufficientPredicates(this, d, insufficientAltToLocations); 
        if (resolvedStates == null || !resolvedStates.contains(d)) {
          Set<Integer> disabledAlts = getDisabledAlternatives(d);
          stripWildCardAlts(disabledAlts);
          if (disabledAlts.size() > 0) {
            boolean explicitlyGreedy = false;
            GrammarAST blockAST = d.dfa.nfa.grammar.getDecisionBlockAST(d.dfa.decisionNumber);
            if (blockAST != null) {
              String greedyS = (String)blockAST.getBlockOption("greedy");
              if (greedyS != null && greedyS.equals("true"))
                explicitlyGreedy = true; 
            } 
            if (!explicitlyGreedy)
              ErrorManager.nondeterminism(this, d); 
          } 
        } 
      } 
    } 
    Set<DFAState> danglingStates = getDanglingStates();
    if (danglingStates.size() > 0)
      for (DFAState d : danglingStates)
        ErrorManager.danglingState(this, d);  
    if (!this.nonLLStarDecision) {
      List<Integer> unreachableAlts = this.dfa.getUnreachableAlts();
      if (unreachableAlts != null && unreachableAlts.size() > 0) {
        boolean isInheritedTokensRule = false;
        if (this.dfa.isTokensRuleDecision())
          for (Integer altI : unreachableAlts) {
            GrammarAST decAST = this.dfa.getDecisionASTNode();
            GrammarAST altAST = (GrammarAST)decAST.getChild(altI.intValue() - 1);
            GrammarAST delegatedTokensAlt = (GrammarAST)altAST.getFirstChildWithType(29);
            if (delegatedTokensAlt != null) {
              isInheritedTokensRule = true;
              ErrorManager.grammarWarning(162, this.dfa.nfa.grammar, null, this.dfa.nfa.grammar.name, delegatedTokensAlt.getChild(0).getText());
            } 
          }  
        if (!isInheritedTokensRule)
          ErrorManager.unreachableAlts(this, unreachableAlts); 
      } 
    } 
  }
  
  protected void stripWildCardAlts(Set<Integer> disabledAlts) {
    GrammarAST lastAltAST;
    List<Integer> sortedDisableAlts = new ArrayList<Integer>(disabledAlts);
    Collections.sort(sortedDisableAlts);
    Integer lastAlt = sortedDisableAlts.get(sortedDisableAlts.size() - 1);
    GrammarAST blockAST = this.dfa.nfa.grammar.getDecisionBlockAST(this.dfa.decisionNumber);
    if (blockAST.getChild(0).getType() == 58) {
      lastAltAST = (GrammarAST)blockAST.getChild(lastAlt.intValue());
    } else {
      lastAltAST = (GrammarAST)blockAST.getChild(lastAlt.intValue() - 1);
    } 
    if (lastAltAST.getType() != 33 && lastAltAST.getChild(0).getType() == 98 && lastAltAST.getChild(1).getType() == 32)
      disabledAlts.remove(lastAlt); 
  }
  
  protected void issueRecursionWarnings() {
    Set<Integer> dfaStatesWithRecursionProblems = this.stateToRecursionOverflowConfigurationsMap.keySet();
    Map<Integer, Map<String, Set<NFAState>>> altToTargetToCallSitesMap = new HashMap<Integer, Map<String, Set<NFAState>>>();
    Map<Integer, DFAState> altToDFAState = new HashMap<Integer, DFAState>();
    computeAltToProblemMaps(dfaStatesWithRecursionProblems, this.stateToRecursionOverflowConfigurationsMap, altToTargetToCallSitesMap, altToDFAState);
    Set<Integer> alts = altToTargetToCallSitesMap.keySet();
    List<Integer> sortedAlts = new ArrayList<Integer>(alts);
    Collections.sort(sortedAlts);
    for (Integer altI : sortedAlts) {
      Map<String, Set<NFAState>> targetToCallSiteMap = altToTargetToCallSitesMap.get(altI);
      Set<String> targetRules = targetToCallSiteMap.keySet();
      Collection<Set<NFAState>> callSiteStates = targetToCallSiteMap.values();
      DFAState sampleBadState = altToDFAState.get(altI);
      ErrorManager.recursionOverflow(this, sampleBadState, altI.intValue(), targetRules, (Collection)callSiteStates);
    } 
  }
  
  private void computeAltToProblemMaps(Set<Integer> dfaStatesUnaliased, Map<Integer, List<NFAConfiguration>> configurationsMap, Map<Integer, Map<String, Set<NFAState>>> altToTargetToCallSitesMap, Map<Integer, DFAState> altToDFAState) {
    for (Integer stateI : dfaStatesUnaliased) {
      List<? extends NFAConfiguration> configs = configurationsMap.get(stateI);
      for (int i = 0; i < configs.size(); i++) {
        NFAConfiguration c = configs.get(i);
        NFAState ruleInvocationState = this.dfa.nfa.getState(c.state);
        Transition transition0 = ruleInvocationState.transition[0];
        RuleClosureTransition ref = (RuleClosureTransition)transition0;
        String targetRule = ((NFAState)ref.target).enclosingRule.name;
        Integer altI = Utils.integer(c.alt);
        Map<String, Set<NFAState>> targetToCallSiteMap = altToTargetToCallSitesMap.get(altI);
        if (targetToCallSiteMap == null) {
          targetToCallSiteMap = new HashMap<String, Set<NFAState>>();
          altToTargetToCallSitesMap.put(altI, targetToCallSiteMap);
        } 
        Set<NFAState> callSites = targetToCallSiteMap.get(targetRule);
        if (callSites == null) {
          callSites = new HashSet<NFAState>();
          targetToCallSiteMap.put(targetRule, callSites);
        } 
        callSites.add(ruleInvocationState);
        if (altToDFAState.get(altI) == null) {
          DFAState sampleBadState = this.dfa.getState(stateI.intValue());
          altToDFAState.put(altI, sampleBadState);
        } 
      } 
    } 
  }
  
  private Set<Integer> getUnaliasedDFAStateSet(Set<Integer> dfaStatesWithRecursionProblems) {
    Set<Integer> dfaStatesUnaliased = new HashSet<Integer>();
    for (Integer stateI : dfaStatesWithRecursionProblems) {
      DFAState d = this.dfa.getState(stateI.intValue());
      dfaStatesUnaliased.add(Utils.integer(d.stateNumber));
    } 
    return dfaStatesUnaliased;
  }
  
  public void reportDanglingState(DFAState d) {
    this.danglingStates.add(d);
  }
  
  public void reportNonLLStarDecision(DFA dfa) {
    this.nonLLStarDecision = true;
    dfa.nfa.grammar.numNonLLStar++;
    this.altsWithProblem.addAll(dfa.recursiveAltSet.toList());
  }
  
  public void reportRecursionOverflow(DFAState d, NFAConfiguration recursionNFAConfiguration) {
    if (d.stateNumber > 0) {
      Integer stateI = Utils.integer(d.stateNumber);
      this.stateToRecursionOverflowConfigurationsMap.map(stateI, recursionNFAConfiguration);
    } 
  }
  
  public void reportNondeterminism(DFAState d, Set<Integer> nondeterministicAlts) {
    this.altsWithProblem.addAll(nondeterministicAlts);
    this.statesWithSyntacticallyAmbiguousAltsSet.add(d);
    this.dfa.nfa.grammar.setOfNondeterministicDecisionNumbers.add(Utils.integer(this.dfa.getDecisionNumber()));
  }
  
  public void reportLexerRuleNondeterminism(DFAState d, Set<Integer> nondeterministicAlts) {
    this.stateToSyntacticallyAmbiguousTokensRuleAltsMap.put(d, nondeterministicAlts);
  }
  
  public void reportNondeterminismResolvedWithSemanticPredicate(DFAState d) {
    if (d.abortedDueToRecursionOverflow)
      d.dfa.probe.removeRecursiveOverflowState(d); 
    this.statesResolvedWithSemanticPredicatesSet.add(d);
    this.dfa.nfa.grammar.setOfNondeterministicDecisionNumbersResolvedWithPredicates.add(Utils.integer(this.dfa.getDecisionNumber()));
  }
  
  public void reportAltPredicateContext(DFAState d, Map<Integer, ? extends SemanticContext> altPredicateContext) {
    Map<Integer, SemanticContext> copy = new HashMap<Integer, SemanticContext>();
    copy.putAll(altPredicateContext);
    this.stateToAltSetWithSemanticPredicatesMap.put(d, copy);
  }
  
  public void reportIncompletelyCoveredAlts(DFAState d, Map<Integer, Set<Token>> altToLocationsReachableWithoutPredicate) {
    this.stateToIncompletelyCoveredAltsMap.put(d, altToLocationsReachableWithoutPredicate);
  }
  
  protected boolean reachesState(DFAState startState, DFAState targetState, Set<DFAState> states) {
    if (startState == targetState) {
      states.add(targetState);
      this.stateReachable.put(Integer.valueOf(startState.stateNumber), REACHABLE_YES);
      return true;
    } 
    DFAState s = startState;
    this.stateReachable.put(Integer.valueOf(s.stateNumber), REACHABLE_BUSY);
    for (int i = 0; i < s.getNumberOfTransitions(); i++) {
      Transition t = s.transition(i);
      DFAState edgeTarget = (DFAState)t.target;
      Integer targetStatus = this.stateReachable.get(Integer.valueOf(edgeTarget.stateNumber));
      if (targetStatus != REACHABLE_BUSY) {
        if (targetStatus == REACHABLE_YES) {
          this.stateReachable.put(Integer.valueOf(s.stateNumber), REACHABLE_YES);
          return true;
        } 
        if (targetStatus != REACHABLE_NO)
          if (reachesState(edgeTarget, targetState, states)) {
            states.add(s);
            this.stateReachable.put(Integer.valueOf(s.stateNumber), REACHABLE_YES);
            return true;
          }  
      } 
    } 
    this.stateReachable.put(Integer.valueOf(s.stateNumber), REACHABLE_NO);
    return false;
  }
  
  protected Set<DFAState> getDFAPathStatesToTarget(DFAState targetState) {
    Set<DFAState> dfaStates = new HashSet<DFAState>();
    this.stateReachable = new HashMap<Integer, Integer>();
    if (this.dfa == null || this.dfa.startState == null)
      return dfaStates; 
    boolean reaches = reachesState(this.dfa.startState, targetState, dfaStates);
    return dfaStates;
  }
  
  protected void getSampleInputSequenceUsingStateSet(State startState, State targetState, Set<DFAState> states, List<Label> labels) {
    this.statesVisitedDuringSampleSequence.add(Integer.valueOf(startState.stateNumber));
    for (int i = 0; i < startState.getNumberOfTransitions(); i++) {
      Transition t = startState.transition(i);
      DFAState edgeTarget = (DFAState)t.target;
      if (states.contains(edgeTarget) && !this.statesVisitedDuringSampleSequence.contains(Integer.valueOf(edgeTarget.stateNumber))) {
        labels.add(t.label);
        if (edgeTarget != targetState)
          getSampleInputSequenceUsingStateSet(edgeTarget, targetState, states, labels); 
        return;
      } 
    } 
    labels.add(new Label(-5));
  }
  
  protected boolean getNFAPath(NFAState s, int labelIndex, List<? extends Label> labels, List<? super NFAState> path) {
    String thisStateKey = getStateLabelIndexKey(s.stateNumber, labelIndex);
    if (this.statesVisitedAtInputDepth.contains(thisStateKey))
      return false; 
    this.statesVisitedAtInputDepth.add(thisStateKey);
    for (int i = 0; i < s.getNumberOfTransitions(); i++) {
      Transition t = s.transition[i];
      NFAState edgeTarget = (NFAState)t.target;
      Label label = labels.get(labelIndex);
      if (t.label.isEpsilon() || t.label.isSemanticPredicate()) {
        path.add(edgeTarget);
        boolean found = getNFAPath(edgeTarget, labelIndex, labels, path);
        if (found) {
          this.statesVisitedAtInputDepth.remove(thisStateKey);
          return true;
        } 
        path.remove(path.size() - 1);
      } else if (t.label.matches(label)) {
        path.add(edgeTarget);
        if (labelIndex == labels.size() - 1) {
          this.statesVisitedAtInputDepth.remove(thisStateKey);
          return true;
        } 
        boolean found = getNFAPath(edgeTarget, labelIndex + 1, labels, path);
        if (found) {
          this.statesVisitedAtInputDepth.remove(thisStateKey);
          return true;
        } 
        path.remove(path.size() - 1);
      } 
    } 
    this.statesVisitedAtInputDepth.remove(thisStateKey);
    return false;
  }
  
  protected String getStateLabelIndexKey(int s, int i) {
    StringBuilder buf = new StringBuilder();
    buf.append(s);
    buf.append('_');
    buf.append(i);
    return buf.toString();
  }
  
  public String getTokenNameForTokensRuleAlt(int alt) {
    NFAState decisionState = this.dfa.getNFADecisionStartState();
    NFAState altState = this.dfa.nfa.grammar.getNFAStateForAltOfDecision(decisionState, alt);
    NFAState decisionLeft = (NFAState)(altState.transition[0]).target;
    RuleClosureTransition ruleCallEdge = (RuleClosureTransition)decisionLeft.transition[0];
    NFAState ruleStartState = (NFAState)ruleCallEdge.target;
    return ruleStartState.enclosingRule.name;
  }
  
  public void reset() {
    this.stateToRecursionOverflowConfigurationsMap.clear();
  }
}
